
// =============================================================
// ===            Algebraic recognition routines             ===
// =============================================================

// Here is a series of routines that are (as far as I know) not implemented in magma.
// We have used the names of the corresponding routines in PARI/GP, which I have hopefully implemented correctly!
        
// ============================================================
// ===                 Table of contents                    ===
// ============================================================

//    (1) General algebraic recognition
//    (2) Algebraic recognition mod eps^Z
//    (3) Algebraic recognition mod Tate(q)^Z

        
// =============================================================
// ===           1) General algebraic recognition            ===
// =============================================================


// Algebraic recognition for element in Z/p^mZ.
function algdepZpm(a,deg)

	N  := Modulus(Parent(a));
	ZZ := Integers();
	RR := RealField(500);
	M  := ZeroMatrix(RR,deg+2,deg+2);

	for i:=0 to deg do
		M[i+1,1] := ZZ!((a)^i);
	end for;
	M[deg+2][1] := N;

	for j:=1 to deg+1 do
		M[j,j+1] := 1;
	end for;

	Y,T := LLL(M);

	PolZ<t> := PolynomialRing(ZZ);
	P       := Floor(Y[2][1] - Y[2][2]);
	for j := 3 to deg+2 do
		P := P - Floor(Y[2][j])*t^(j-2);
	end for;

	return P;

end function;

// Algebraic recognition for element in Qp.
function algdepQp(a,deg)

    ZZ := Integers();
    QQ := Rationals();
    RR := RealField(500);
    PolZ<x>:=PolynomialRing(ZZ);
   
    p := Prime(Parent(a));
    m := Precision(Parent(a));
    N := p^m;

    M:=ZeroMatrix(RR,deg+2,deg+2);
    for j := 1 to deg+1 do 
        M[j,j] := 1;
        M[j,deg+2] := QQ!(a^(j-1));
    end for;
    M[deg+2][deg+2] := N;
    Y,T:=LLL(M);

    P1 := Floor(Y[2][deg+2] - Y[2][1]);
    for j := 2 to deg+1 do
        P1 := P1 - Floor(Y[2][j])*x^(j-1);
    end for;
    Fac1 := Factorisation(P1);
    P1   := Fac1[#Fac1][1];
    
    return P1;
       
end function;

// Algebraic recognition for element in Qp^2.
function algdepQp2(a,deg)

    ZZ := Integers();
    QQ := Rationals();
    RR := RealField(500);
    PolZ<x>:=PolynomialRing(ZZ);
   
    p := Prime(Parent(a));
    m := Precision(Parent(a));
    N := p^Floor(5*m/7);

    M:=ZeroMatrix(RR,deg+3,deg+3);
    for j := 1 to deg+1 do 
        M[j,j] := 1;
        M[j,deg+2] := QQ!(Coefficient((a)^(j-1),1));
        M[j,deg+3] := QQ!(Coefficient((a)^(j-1),2));
    end for;
    M[deg+2][deg+2] := N;
    M[deg+3][deg+3] := N;
    Y,T:=LLL(M);

    P1 := Floor(Y[2][deg+2] - Y[2][1]);
    for j := 2 to deg+1 do
        P1 := P1 - Floor(Y[2][j])*x^(j-1);
    end for;
    Fac1 := Factorisation(P1);
    P1   := Fac1[#Fac1][1];
    
    return P1;
       
end function;


// Given a number a, attempt to recognise it as an algebraic number.
function algdep(a,deg)
	
	if Type(a) eq RngIntResElt then
		return algdepZpm(a,deg);
	else
		if Degree(Parent(a)) eq 1 then
			return algdepQp(a,deg);
		else
			return algdepQp2(a,deg);
		end if;
	end if;	

end function;


           
// Given a list of numbers in Qp^2, find an integer linear relation between them
function lindep(List : prt := true)

    ZZ := Integers();
    QQ := Rationals();
    RR:=RealField(500);
    
    Kp := Parent(List[2]);
    p := Prime(Kp);
    m := Precision(Kp);
    N := p^(Floor(5*m/7));
    
    len := #List;
    M := ZeroMatrix(RR,len+2,len+2);
    for j:=1 to len do
        M[j,j]:=1;
        M[j,len+1]:=QQ!(Coefficient(List[j],1));
        M[j,len+2]:=QQ!(Coefficient(List[j],2));
    end for;
    M[len+1][len+1]:=N;
    M[len+2][len+2]:=N;
    Y,T:=LLL(M);
    
    //print ChangeRing(RowSubmatrix(Y,2),Integers());
    //print "";
    
	vec := [ZZ!(Y[2][len+1] - Y[2][1])];
	sum := vec[1];
	for i := 2 to len do
		Append(~vec,ZZ!(-Y[2][i]));
		sum := sum + vec[i]*List[i];
	end for;
	
	if prt eq true then
		print "If lindep succeeded, this should be zero: ", sum; 
	end if;

    return vec, sum;
   
end function;


// Prints the Newton slopes at all primes dividing the constant term of P1. 
function ValuationVectors(P1)

	ZZ  := Integers();
	K   := Parent(Coefficient(P1,0));
	Fac := Factorisation(Coefficient(P1,0)*Integers(K));
	qprev := 1;
	ind   := 1;
	
	for Q in Fac do
		q := Q[1];
		qrat := Factorisation(NormAbs(q))[1][1];
		if qrat eq qprev then
			ind := ind +1;
		else
			ind := 1;
			qprev := qrat;
		end if;
		if Degree(K) eq 1 then
			V := Vertices(NewtonPolygon(P1,Generator(q)));
		else
			V := Vertices(NewtonPolygon(P1,q));
		end if;
		if #V ne 0 then
			Vecq := [];
			for i := 1 to #V-1 do
				mult  := V[i+1][1] - V[i][1];
				slope := (V[i][2] - V[i+1][2])/mult;
				for j := 1 to ZZ!mult do
					Append(~Vecq,slope);
				end for;
			end for;
			for i := 1 to Floor(Degree(P1)/2)-ZZ!(V[#V][1]) do
				Append(~Vecq,0);
			end for; 
			for j := 1 to Floor(Degree(P1)/2) do
				Append(~Vecq,-Vecq[Floor(Degree(P1)/2)+1-j]);
			end for;
			print " ";
			if Degree(K) gt 1 then
				print "Newton slopes at prime", ind, " above ", qrat, ":";
			else
				print "Newton slopes at", qrat, ":";
			end if;
			print Vecq;
		end if;
	end for;
	
	return " ";

end function;


// =============================================================
// ===          2) Algebraic recognition mod eps^Z           ===
// =============================================================
   

// Returns height of the polynomial f.
function PolyHeight(f)

    h := 0;
    for i := 0 to Degree(f) do
        ci := Abs(Coefficient(f,i));
        if ci gt h then
            h := ci;
        end if;
    end for;
   
    return h;

end function;

// Find small height integer polynomial of degree d which has A as a root modulo m.
function ThoroughAlgdepEpsilon(A,deg,DF:depth := 120)

	Kp := Parent(A);
	p := Prime(Parent(A));
	m := Precision(Parent(A));
	N := p^m;
			
	OO   := QuadraticOrder(QuadraticForms(DF));
	feps := MinimalPolynomial(FundamentalUnit(OO));
	eps  := Roots(feps,Kp)[1][1];
	zeta := Roots(CyclotomicPolynomial(p^2-1),Kp)[1][1];
	
	ZZ := Integers();
	QQ := Rationals();
	RR := RealField(500);
	
	PolZ<x>:=PolynomialRing(ZZ);
	Currdeg := deg;
	CurrP := 10^200*x^deg;
	CurrHt:= PolyHeight(CurrP);
	Currk := -depth;
	Currl := 0;
	
	n := depth; // DEPTH of the algorithm, increase if necessary!
	for k := -n to n do
		for l := 0 to p do
			a := A*zeta^l*eps^k;
			P1 := algdep(a,deg);			
        	if Valuation(Evaluate(P1,a)) gt Floor(5*m/7) then
            	if Degree(P1) lt Currdeg or PolyHeight(P1) lt CurrHt then
                	Currdeg := Degree(P1);
                	CurrP := P1;
                	CurrHt:= PolyHeight(P1);
                	Currk := k;
                	Currl := l;
            	end if;
        	end if;
		end for;
	end for;
	
	a := A*zeta^Currl*eps^Currk;	
	if Valuation(Evaluate(CurrP,a)) gt Floor(5*m/7) and CurrP ne 10^200*x^deg then
        return CurrP, a;
    else
//     	print "Not recognised!";
        return 1*x^0, a;
    end if;
	
end function;

// Find small height integer polynomial of degree d which has A as a root modulo m.
function ThoroughAlgdepZeta(a,deg)
   
    Kp   := Parent(a);
    p    := Prime(Kp);
    m    := Precision(Parent(a));
    zeta := Roots(CyclotomicPolynomial(p^2-1),Kp)[1][1];
   
    ZZ := Integers();
    QQ := Rationals();
    RR := RealField(500);
   
    PolZ<x> := PolynomialRing(ZZ);
    Currdeg := deg;
    CurrP := 10^200*x^deg;
    CurrHt:= PolyHeight(CurrP);
    Currk := 0;
   
    for k := 0 to p^2-2 do
        A := a*zeta^k;
        P1 := algdep(A,deg);
        if Valuation(Evaluate(P1,A)) gt Floor(5*m/7) then
            if Degree(P1) lt Currdeg or PolyHeight(P1) lt CurrHt then
                Currdeg := Degree(P1);
                CurrP := P1;
                CurrHt:= PolyHeight(P1);
                Currk := k;
            end if;
        end if;
    end for;
        
    A := a*zeta^Currk;	
	if Valuation(Evaluate(CurrP,A)) gt Floor(5*m/7) and CurrP ne 10^200*x^deg then
        return CurrP, A;
    else
//    	print "Not recognised!";
        return 1*x^0, A;
    end if;

end function;



// =============================================================
// ===        3) Algebraic recognition mod Tate(q)^Z         ===
// =============================================================
   
// Magma does not seem to have the same functionality as Sage when it comes to Tate curves. 
// Here are some routines we need. 

// This function computes the q-parameter of a Tate curve E, to precision m.
function qParameter(E,p,m)

	jE   := jInvariant(E);
	vjE  := Valuation(jE,p);
	assert vjE lt 0; // E must be a Tate curve!
	
	Qp   := pAdicField(p,m);
	R<q> := PowerSeriesRing(Qp,m);
	Jq   := jInvariant(q);
	g    := Reverse(1/Jq);
	
	return Evaluate(g,1/jE);

end function;


// This function computes the sum sk(q), see Silverman p.423
function TateSk(q,k,m)

	Sk := q/(1-q);
	qi := q;
	for i := 2 to m-1 do
		qi := qi*q;
		Sk := Sk + i^k*qi/(1-qi);
	end for;
	
	return Sk;
	
end function;

// This function computes the coefficient a4(q), see Silverman p.423
function TateA4(q,m)

	return -5*TateSk(q,3,m);
	
end function;

// This function computes the coefficient a4(q), see Silverman p.423
function TateA6(q,m)

	ZZ := Integers();
	A6 := 0;
	qi := 1;
	for i := 1 to m-1 do
		qi := qi*q; 
		A6 := A6 + -ZZ!((5*i^3+7*i^5)/12)*qi/(1-qi);
	end for;
	
	return A6;
	
end function;

// This function returns the X-coordinate of the point u in Cp*/q^Z, up to precision m.
function TateX(u,q)

	m := Precision(Parent(u));
	X := u/(1-u)^2;
	qi := 1;
	for i := 1 to m-1 do
		qi := qi*q;
		X := X + qi*u/(1-qi*u)^2 + qi*u/(u-qi)^2 - 2*qi/(1-qi)^2;
	end for;
	
	return X;

end function;

// This function returns the Y-coordinate of the point u in Cp*/q^Z, up to precision m.
function TateY(u,q)

	m := Precision(Parent(u));
	Y := u^2/(1-u)^3;
	qi := 1;
	for i := 1 to m-1 do
		qi := qi*q;
		Y := Y + (qi*u)^2/(1-qi*u)^3 - qi*u^2/(u-qi)^3 + qi/(1-qi)^2;
	end for;
	
	return Y;

end function;
