// ============================================================
// ===                SAMPLE QUADRATIC FORMS                ===
// ============================================================

// A list of all quadratic forms up to discriminant 1000.
// Naming convention is: All forms of discriminant D are labelled D(disc)F(i), where "disc" is the discriminant, and "i" ranges from 1 to the narrow class number.
// For instance, in discriminant 44 there are precisely two forms, named D44F1 and D44F2.

load "Discriminants.m";

// ============================================================
// ===               FREQUENTLY USED MATRICES               ===
// ============================================================

IMat := Matrix(2,2,[[1, 0],[0,1]]);  // Identity
SMat := Matrix(2,2,[[0,-1],[1,0]]);  // Order 4
TMat := Matrix(2,2,[[1, 1],[0,1]]);  // Translation
STMat:= Matrix(2,2,[[0,-1],[1,1]]);  // Order 6 

// ============================================================
// ===                 Table of contents                    ===
// ============================================================

//    (1) General Quadratic Form Routines
//    (2) Nearly reduced cycles
//    (3) Power series in level 0 and 1
//    (4) Tp and continued fractions
//    (5) Operations on power series
//    (6) Main functions
//    (7) Writing functions
//    (8) Tests

// ============================================================
// ===          1) General Quadratic Form Routines          ===
// ============================================================

// Find the (ordered) roots of a quadratic form, as an element in a degree 2 ext of Q.
function FormRoots(Q,K)
	
	S<x> := PolynomialRing(K);
	f := Q[1]*x^2+Q[2]*x+Q[3];
	Rts := Roots(f);
	return [Minimum(Rts)[1],Maximum(Rts)[1]];

end function;

// Finds all the "genus 0" primes p=2,3,5,7,13 which are inert in the order of quadratic form F
function InertGenus0Primes(F)

	DF := Discriminant(F);
	P := [];
	C := [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 41, 47, 59, 71];
	for p in C do
		if KroneckerSymbol(DF,p) eq -1 then
			Append(~P,p);
		end if;
	end for;
	
	return P;

end function;

// Finds all the "genus 0" primes p=2,3,5,7,13 which are split in the order of quadratic form F
function SplitGenus0Primes(F)

	DF := Discriminant(F);
	P := [];
	C := [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 41, 47, 59, 71];
	for p in C do
		if KroneckerSymbol(DF,p) eq 1 then
			Append(~P,p);
		end if;
	end for;
	
	return P;

end function;

// Check whether a form is primitive.
function IsPrimitiveForm(Q)

	return Gcd(Gcd(Q[1],Q[2]),Q[3]) eq 1;

end function;

// Check whether a form is primitive.
function PrimitiveForm(Q)

	n := Gcd(Gcd(Q[1],Q[2]),Q[3]);
	D := Discriminant(Q);
	ZZ := Integers();

	return QuadraticForms(ZZ!(D/n^2))!<ZZ!(Q[1]/n),ZZ!(Q[2]/n),ZZ!(Q[3]/n)>;

end function;

// Given F = Ax^2 + Bxy + Cy^2, return Finv = Ax^2 - Bxy + Cy^2
function AmbiguousForm(F)
	
	DF := Discriminant(F);
	QQ := QuadraticForms(DF);
	
	return QQ![F[1],-F[2],F[3]]; 

end function;

// Given F = Ax^2 + Bxy + Cy^2, return Finv = -Ax^2 - Bxy - Cy^2
function ReciprocalForm(F)
	
	DF := Discriminant(F);
	QQ := QuadraticForms(DF);
	
	return QQ![-F[1],-F[2],-F[3]]; 

end function;

// Given F = Ax^2 + Bxy + Cy^2, return Finv = -Ax^2 + Bxy - Cy^2
function AntiguousForm(F)
	
	DF := Discriminant(F);
	QQ := QuadraticForms(DF);
	
	return QQ![-F[1],F[2],-F[3]]; 

end function;

// Finds all discriminants that can be paired up with the p-adic theta series of F (discriminant DF).
function InertDiscriminants(p,I:DF := -1)

	L := [];
	for D in I do
		if IsDiscriminant(D) and KroneckerSymbol(FundamentalDiscriminant(D),p) eq -1 and D ne DF then
			Append(~L,D);
		end if;
	end for;
	
	return L;		

end function;

// Finds all discriminants that can be paired up with the p-adic theta series of F (discriminant DF).
function SplitDiscriminants(p,I:DF := -1)

	L := [];
	for D in I do
		if IsDiscriminant(D) and KroneckerSymbol(FundamentalDiscriminant(D),p) eq 1 and D ne DF then
			Append(~L,D);
		end if;
	end for;
	
	return L;		

end function;



function DerivedForms(QQ,p)

	prim    := [];
	nonprim := [];
	
	DQ := Discriminant(QQ);
	a := QQ[1];
	b := QQ[2];
	c := QQ[3];
	for j := 0 to p-1 do
		QQjp := QuadraticForms(DQ*p^2)!<a*p^2,p*(b+2*j*a),a*j^2+b*j+c>;
		if IsPrimitiveForm(QQjp) then
			Append(~prim,QQjp);
		else
			Append(~nonprim,PrimitiveForm(QQjp));
		end if;
	end for;
		
	QQinfp := QuadraticForms(DQ*p^2)!<c*p^2,-b*p,a>;
	if IsPrimitiveForm(QQinfp) then
		Append(~prim,QQinfp);
	else
		Append(~nonprim,PrimitiveForm(QQinfp));
	end if;
	
	return prim, nonprim;

end function;


function pDerivedTest(QQ,p,n)

	assert KroneckerSymbol(Discriminant(QQ),p) eq 1;

	classes0 := {@ ReductionOrbit(QQ) @};
	classes1 := {@ @};

	for i := 1 to n do
		prim, nonprim := DerivedForms(classes0[#classes0][1],p);
		print "The two non-primitive forms are equivalent: ", IsEquivalent(nonprim[1],nonprim[2]);
		classes1 := classes1 join {@ ReductionOrbit(nonprim[1]) @} join {@ ReductionOrbit(nonprim[2]) @};
		prim, nonprim := DerivedForms(classes1[#classes1][1],p);
		print "The two non-primitive forms are equivalent: ", IsEquivalent(nonprim[1],nonprim[2]);
		classes0 := classes0 join {@ ReductionOrbit(nonprim[1]) @} join {@ ReductionOrbit(nonprim[2]) @};
	end for;
	
	return classes0, classes1;

end function;


// ============================================================
// ===              2) Nearly reduced cycles                ===
// ============================================================

// Given QQ, this function returns all SL_2(Z)-translates Q' of Q whose roots are separated by 0.
// This means that one of the roots of Q' is less than 0, and the other is greater than 0.
function NearlyReducedForms(QQ)

	DQ := Discriminant(QQ);
	KK := QuadraticField(DQ);

	SepForms := [];
	Red := ReductionOrbit(QQ);

	for Q in Red do
		RtsQ := FormRoots(Q,KK);
		A := Floor(RtsQ[1])+1;
		B := Floor(RtsQ[2]);
		
		for i := A to B do
			Mi := Matrix([[1,i],[0,1]]);
			QNear := Q*Mi;
			Append(~SepForms,QNear);

			if i ne A and i ne B then
				QNear := Q*Mi*SMat;
				Append(~SepForms,QNear);
			end if;
		end for;
	end for;
		
	return SepForms;

end function;

// Determines j = 0,1,...,p-1,infty such that r = j mod p.
function ResidueClass(r,p)

	A := Denominator(r);
	if IsDivisibleBy(A,p) then
		return p;
	else
		Fp := GF(p);
		return Integers()!(Fp!r);
	end if;
	
end function;

// Determines j = 0,1,...,p-1,infty such that the roots of Q are j mod p.
// WARNING: We assume Q is "Level 1", meaning v_p(Disc(Q)) = 2.
function ResidueClassForm(Q,p)

	if Valuation(Q[1],p) eq 0 then
		Rts := FormRoots(Q,GF(p));
		assert Rts[1] eq Rts[2];
		return Integers()!Rts[1];
	else
		return p;
	end if;

end function;

// Tests whether a form QQ is equivalent to any of the forms in the list L.
function IsEquivalentList(QQ,L)

	a := false;
	for F in L do
		if IsEquivalent(QQ,F) then
			a := true;
		end if;
	end for;
	
	return a;
	
end function;


// Finds all the forms in level 0 equivalent to Q under Gammatilde.
function GammatildeLevel0(Q,DQ,p)

	Forms := [* [Q], [] *];;
	if KroneckerSymbol(DQ,p) eq -1 then
		return Forms;
	else
		h := #ReducedOrbits(QuadraticForms(DQ));
		for i := 1 to h do
			// p-derived forms from Gamma orbit
			for QQ in Forms[1] do
				prim, nonprim := DerivedForms(QQ,p);
				for FF in nonprim do
					F := PrimitiveForm(FF);
					if IsEquivalentList(F,Forms[2]) eq false then
						Append(~Forms[2],F);
					end if;
				end for;
			end for;
			// p-derived forms from Gammatilde \ Gamma orbit
			for QQ in Forms[2] do
				prim, nonprim := DerivedForms(QQ,p);
				for FF in nonprim do
					F := PrimitiveForm(FF);
					if IsEquivalentList(F,Forms[1]) eq false then
						Append(~Forms[1],F);
					end if;
				end for;
			end for;
		end for;
		
		return Forms; 
		
	end if;		

end function;

// Given Q, this function returns all SL_2(Z)-translates Q' of Q that are separated by 0.
// This means that one of the roots of Q' is less than 0, and the other is greater than 0.
// They are returned as a list, indexed by which residue disk its roots belong to.

// OUTPUT:
// SepForms0 = [* S0even, S0odd *]      -      The nearly reduced forms Gammatilde-equivalent to QQ
// SepForms1 = [* S1even, S1odd *]      -      The nearly reduced forms Gammatilde-equivalent to one of the forms
//                                                      QQ | [[1 0] [0 p]] or QQ | [[p j] [0 1]]
//                                             (There are p+1 sequences, one for every residue disk in which their roots lie)

function NearlyReducedFormsLevel01(Q,p)

	DQ := Discriminant(Q);
	KK := QuadraticField(DQ);
	
	// STEP 1: Find all nearly reduced forms of level 0 in Gammatilde orbit of Q.
    // LEVEL 0:	
    Forms0 := GammatildeLevel0(Q,DQ,p);
    // Gamma
    SepForms0even := [];
    for QQ in Forms0[1] do
    	SepForms0even := SepForms0even cat NearlyReducedForms(QQ);
    end for;
    // Gammatilde \ Gamma
    SepForms0odd := [];
    for QQ in Forms0[2] do
    	SepForms0odd := SepForms0odd cat NearlyReducedForms(QQ);
    end for; 
	
	// STEP 2: Find all reduced forms of level 1 in Gammatilde orbit
	
	Red1even := {@ @};
	Red1odd  := {@ @};
	
	// Reduced forms, Gammma.
	for QQ in Forms0[1] do
		a := QQ[1];
		b := QQ[2];
		c := QQ[3];
		for j := 0 to p-1 do
			QQjp := QuadraticForms(DQ*p^2)!<a*p^2,p*(b+2*j*a),a*j^2+b*j+c>;
			// If p is inert, this form is always primitive, but if p is split
			// there are precisely 2 imprimitive forms that arise this way, and we take them out.
			if IsPrimitiveForm(QQjp) then
				Red1odd := Red1odd join ReductionOrbit(QQjp);
			end if;
		end for;
		QQinfp := QuadraticForms(DQ*p^2)!<c*p^2,-b*p,a>;
		if IsPrimitiveForm(QQinfp) then
			Red1odd := Red1odd join ReductionOrbit(QQinfp);
		end if;
	end for;
	
	// Reduced forms, Gammmatilde \ Gamma.
	for QQ in Forms0[2] do
		a := QQ[1];
		b := QQ[2];
		c := QQ[3];
		for j := 0 to p-1 do
			QQjp := QuadraticForms(DQ*p^2)!<a*p^2,p*(b+2*j*a),a*j^2+b*j+c>;
			// If p is inert, this form is always primitive, but if p is split
			// there are precisely 2 imprimitive forms that arise this way, and we take them out.
			if IsPrimitiveForm(QQjp) then
				Red1even := Red1even join ReductionOrbit(QQjp);
			end if;
		end for;
		QQinfp := QuadraticForms(DQ*p^2)!<c*p^2,-b*p,a>;
		if IsPrimitiveForm(QQinfp) then
			Red1even := Red1even join ReductionOrbit(QQinfp);
		end if;
	end for;
			
	// STEP 3: Now find nearly reduced orbit from set of all reduced forms found above.
	
	SepForms1even := [[] : j in [0..p]];
	SepForms1odd  := [[] : j in [0..p]];				
			
	// Nearly reduced forms, Gamma.
	for Q in Red1even do
		RtsQ := FormRoots(Q,KK);
		A := Floor(RtsQ[1])+1;
		B := Floor(RtsQ[2]);
		
		for i := A to B do
			Mi := Matrix([[1,i],[0,1]]);
			QNear := Q*Mi;
			j := ResidueClassForm(QNear,p);
			Append(~SepForms1even[j+1],QNear);

			if i ne A and i ne B then
				QNear := Q*Mi*SMat;
				j := ResidueClassForm(QNear,p);
				Append(~SepForms1even[j+1],QNear);
			end if;
		end for;
	end for;
	
	// Nearly reduced forms, Gammatilde \ Gamma.
	for Q in Red1odd do
		RtsQ := FormRoots(Q,KK);
		A := Floor(RtsQ[1])+1;
		B := Floor(RtsQ[2]);
		
		for i := A to B do
			Mi := Matrix([[1,i],[0,1]]);
			QNear := Q*Mi;
			j := ResidueClassForm(QNear,p);
			Append(~SepForms1odd[j+1],QNear);

			if i ne A and i ne B then
				QNear := Q*Mi*SMat;
				j := ResidueClassForm(QNear,p);
				Append(~SepForms1odd[j+1],QNear);
			end if;
		end for;
	end for;

	return [* SepForms0even,SepForms0odd *], [* SepForms1even,SepForms1odd *];

end function;

// ============================================================
// ===           3) Power series in level 0 and 1           ===
// ============================================================

// OUTPUT: 
// F0       -        F^(0)(0,infty)(z) = Prod_{F'}  (z - w1)/(z - w2) 
// 						where F' ranges over the SL_Z translates of F that separate 0, and w1,w2 are the roots of F' numbered s.t. w1>0 and w2<0
// 						This is computed over the p-adic field Qp^2.
// Fs       -        The sequence of p power series [F_0(t0), F_1(t1), ..., F_oo(t_oo)] of disc Dp^2.
// 						These series are defined as F^(1)_j(0,infty)(z) = Prod_{F'}  (z - w1)/(z - w2), 
//      				where F' ranges over the SL_Z translates of F that separate 0, whose roots w1,w2 are congruent to j mod p, numbered s.t. w1>0 and w2<0
//      				This is computed over the p-adic field Qp^2.
//     					Returned in function of the variable x = 1/(z-j).

function level01series(L0,L1,p,K,sqrtDF,R,Zp,sqrtDFp,Rp : nrm := false)
	
	spl := false;
	if Type(Zp) eq RngIntRes then
		spl := true;
	end if;

	// LEVEL 0:
	z  := R.1;
	F0 := R!1;
	for Q in L0 do
		if Q[1] gt 0 then
			if nrm then
				F0 := F0*(z^2 + Q[2]/Q[1]*z + Q[3]/Q[1])/(z^2 - Q[2]/Q[3]*z + Q[1]/Q[3]);
			else
				r1 := (-Q[2]+sqrtDF)/(2*Q[1]);
				// Split cocycle.
				if spl eq true then
					r2 := (-Q[2]-sqrtDF)/(2*Q[1]);
					F0 := F0*((z - r1)*(z + 1/r2))/((z + 1/r1)*(z - r2));
				// Inert cocycle.
				else
					F0 := F0*(z - r1)/(z + 1/r1);
				end if;
			end if;
		end if;
	end for;
		
	// LEVEL 1:
	Fs := [];
	x := Rp.1;
	for j := 0 to p-1 do
		Fj := Rp!1;
		for Q in L1[j+1] do
			if nrm then
				Fj := Fj*(1+Zp!(Q[2]/Q[1]+2*j)*x+Zp!(Q[3]/Q[1]+j*Q[2]/Q[1]+j^2)*x^2)^(Sign(Q[1]));
			else
				r1 := Zp!((-Q[2]+p*sqrtDFp)/(2*Q[1]));
				// Split cocycle.
				if spl eq true then
					r2 := Zp!((-Q[2]-p*sqrtDFp)/(2*Q[1]));
					Fj := Fj*((1 - (r1-j)*x)/(1 - (r2-j)*x))^(Sign(Q[1]));
				// Inert cocycle.
				else
					Fj := Fj*(1 - (r1-j)*x)^(Sign(Q[1]));
				end if;
			end if;
		end for;
		Append(~Fs,Fj);
	end for;
	
	// Now do F_oo, a power series in t_oo = z
	Foo := Rp!1;
	for Q in L1[p+1] do
		if nrm then
			Foo := Foo*(Zp!(Q[1]/Q[3])*x^2 + Zp!(Q[2]/Q[3])*x + 1)^(Sign(Q[1]));
		else
			r1inv := Zp!((-Q[2]-p*sqrtDFp)/(2*Q[3])); // Inverse of the first root, notice minus sign.
			// Split cocycle.
			if spl eq true then
				r2inv := Zp!((-Q[2]+p*sqrtDFp)/(2*Q[3]));
				Foo   := Foo*((x*r1inv - 1)/(x*r2inv - 1))^(Sign(Q[1]));
			// Inert cocycle.
			else
				Foo   := Foo*(x*r1inv - 1)^(Sign(Q[1]));
			end if;
		end if;
	end for;
	Append(~Fs,Foo);
		
	return [*F0, Fs*];

end function;


// This function returns the level 0 and 1 contributions
//
//     [*    [* F0, F1s *],
//           [* G0, G1s *]    *]
//
// where the Fs are the functions whose roots and poles lie in Gamma-orbit of the roots of F,
// and the Gs are the functions whose roots and poles lie in Gamma_tilde-orbit \ Gamma-orbit of the roots of F.
//
// Note: If p is inert in the field K, then both F1s and G0 are trivial. 
// Hence this data structure refines the odd-even separation of a previous implementation.

function Level01Series(F,p,m,K,sqrtDF,R,Zp,sqrtDFp,Rp : nrm := false)

	L0, L1 := NearlyReducedFormsLevel01(F,p);
	Fs := level01series(L0[1],L1[1],p,K,sqrtDF,R,Zp,sqrtDFp,Rp : nrm := nrm);
	Gs := level01series(L0[2],L1[2],p,K,sqrtDF,R,Zp,sqrtDFp,Rp : nrm := nrm);
	
	return Fs, Gs;
	
end function;

// ============================================================
// ===             4) Tp and continued fractions            ===
// ============================================================


// Partial quotients in continued fraction expansion of rational r.
function PartialQuotients(r)

	CC := ContinuedFraction(r);
	d := #CC;
	if d eq 1 then
		return CC,[1],d;
	else
		P := [CC[1],CC[2]*CC[1]+1];
		Q := [1,CC[2]];
		for i in [3..d] do
			Ci := CC[i];
			Append(~P,Ci*P[i-1]+P[i-2]);
			Append(~Q,Ci*Q[i-1]+Q[i-2]);
		end for;	
		
		return P,Q,d;
	end if;
	
end function;

// Returns a list of matrices in SL_2(Z) whose corresponding paths connect r to infinity.
function GammasUnimodular(r)

	if Type(r) eq Type(Infinity) then
  		return [];
  	else
		P,Q,d := PartialQuotients(r);	
		List := [Matrix(2,2,[1,P[1],0,Q[1]])^(-1)];
	
		for i in [1..d-1] do
			Append(~List,Matrix(2,2,[P[i],(-1)^i*P[i+1],Q[i],(-1)^i*Q[i+1]])^(-1));
		end for;
	
		return List;
	end if;
	
end function;

// For r in Q, finds the rational number Gamma*r.
function MatrixAction(Gamma,r)

	if Type(r) eq Type(Infinity) then
		Gammar := Gamma[1,1]/Gamma[2,1];
	else
		if Gamma[2,1]*r+Gamma[2,2] eq 0 then
			Gammar := Infinity;
		else
			Gammar := (Gamma[1,1]*r+Gamma[1,2])/(Gamma[2,1]*r+Gamma[2,2]);
		end if;
	end if;
	
	return Gammar;

end function;

// Left action of a matrix Gamma on a rational number r, mod p.
// Returns p if infinity.
function MatrixActionModp(Gamma,r,p)

	if (Gamma[2,1]*r+ Gamma[2,2]) eq 0 then
		return p;
	else
		r := (Gamma[1,1]*r + Gamma[1,2])/(Gamma[2,1]*r+ Gamma[2,2]);
		return ResidueClass(r,p);
	end if;
	
end function;


// ============================================================
// ===            5) Operations on power series             ===
// ============================================================

// Multiplies all the series constructed level by level, using the recursion.
function MultiplySeries(Fs,p)

	FF := Fs[2];
	for i := 3 to #Fs do
		for j := 0 to p do
			FF[j+1] := FF[j+1]*Fs[i][j+1];
		end for;
	end for;
			
	return [*Fs[1], FF*];

end function;


// Multiply two lists, the main data structure for our Theta function.
function MultiplyList(L1, L2)

	return [* L1[1]*L2[1], [L1[2][i]*L2[2][i] : i in [1..#L1[2]]] *];

end function;


// Divide two lists, the main data structure for our Theta function.
function DivideList(L1, L2)

	return [* L1[1]/L2[1], [L1[2][i]/L2[2][i] : i in [1..#L1[2]]] *];

end function;

// Divide two lists, the main data structure for our Theta function.
function PowerList(L,n)

	if n ge 0 then
		return [* L[1]^n, [L[2][i]^n : i in [1..#L[2]]] *];
	else
		return [* 1/L[1]^(-n), [1/L[2][i]^(-n) : i in [1..#L[2]]] *];
	end if;

end function;


// Normalise list to have leading terms 1.
function NormaliseList(L)

	return [* L[1]/LeadingCoefficient(Numerator(L[1])), [L[2][i]/Coefficient(L[2][i],0) : i in [1..#L[2]]] *];

end function;


// Computes the n-th power of a cocycle Phi.
function CocyclePower(Phi,n)

	return [* PowerList(Phi[1],n), PowerList(Phi[2],n), Phi[3], Phi[4], Phi[5], Phi[6], Phi[7] *];

end function;


// Computes the product of two cocycles Phi1 and Phi2.
function CocycleProduct(Phi1,Phi2)

	p := Phi1[6];
	m := Phi1[7];
	
	assert p eq Phi2[6]; // Incompatible primes. 
	assert m eq Phi2[7]; // Incompatible precisions.
	
	F1 := MultiplyList(Phi1[1],Phi2[1]);
	F2 := MultiplyList(Phi1[2],Phi2[2]);

	return [* F1, F2, Phi1[3], Phi1[4], Phi1[5], Phi1[6], Phi1[7] *];

end function;

// Computes the quotient of two cocycles Phi1 and Phi2.
function CocycleDivision(Phi1,Phi2)

	p := Phi1[6];
	m := Phi1[7];
	
	assert p eq Phi2[6]; // Incompatible primes. 
	assert m eq Phi2[7]; // Incompatible precisions.
	
	F1 := DivideList(Phi1[1],Phi2[1]);
	F2 := DivideList(Phi1[2],Phi2[2]);

	return [* F1, F2, Phi1[3], Phi1[4], Phi1[5], Phi1[6], Phi1[7] *];

end function;


// Substitution function on cocycles.
function NormaliseCocycle(Phi)
	
	return [* NormaliseList(Phi[1]),NormaliseList(Phi[2]),Phi[3],Phi[4],Phi[5],Phi[6],Phi[7] *];

end function;


// INPUT: 
// Fs      -      A previously computed list whose elements are the level 0 function and the (p+1)-tuples of series computed above.
// gammas  -      A list of matrices in SL_2(Z)
//
// OUTPUT:
// Ls      -      A level 0 function and a (p+1)-tuple of power series, the product 
//                      Ls[j]  =  Prod_{g in gammas} Prod_{t >= 1}  (Fs[t])[j](g*z)

function SubstitutionProduct(Fs,gammas)

	// Global universes
	R  := Parent(Fs[1]);
	z  := R.1;
	// Local universes
	p  := #Fs[2]-1;
	Rp := Parent(Fs[2][1]);
	x  := Rp.1;
	Kp := Parent(Coefficient(Fs[2][1],0));	
	
	Ls := [];
	for j := 0 to p do
		Append(~Ls,Parent(Fs[2][1])!1);
	end for;
	
	// Substitutions.
	f0 := 1;
	for gamma in gammas do
		a := gamma[1,1];
		b := gamma[1,2];
		c := gamma[2,1];
		d := gamma[2,2];
		
		f0 := f0*Evaluate(Fs[1],(a*z+b)/(c*z+d));
		LSub := []; 
		// Finite factors.
		for j := 0 to p-1 do
			gamj := MatrixActionModp(gamma,j,p);
			if gamj eq p then
				sub := (a + (b+a*j)*x)*(c + (d+c*j)*x)^(-1);
			else 
				sub := (c + (d+c*j)*x)*(a + (b+a*j)*x - gamj*(c + (d+c*j)*x))^(-1);
			end if;
			Append(~LSub,Evaluate(Fs[2][gamj+1],sub));
		end for;
		// Infinite factors.
		if c eq 0 then
			gamoo := p;
		else
			gamoo := ResidueClass(a/c,p);
		end if;

		if gamoo eq p then
			sub := (a*x + b)*(c*x + d)^(-1);
		else 
			sub := (c*x+d)*(a*x+b-gamoo*(c*x+d))^(-1);
		end if;
		Append(~LSub,Evaluate(Fs[2][gamoo+1],sub));
	
		for i := 1 to #Ls do
			Ls[i] := Ls[i]*LSub[i];
		end for;
		
	end for;

 	return [* f0, Ls *];

end function;

// Substitution function on cocycles.
function SubstitutionCocycle(Phi,gammas)

	F1 := SubstitutionProduct(Phi[1],gammas);
	F2 := SubstitutionProduct(Phi[2],gammas);

	return [* F1, F2, Phi[3], Phi[4], Phi[5], Phi[6], Phi[7] *];  

end function;


// ============================================================
// ===                  6) Main function                    ===
// ============================================================

// Compute once and for all the powers of the power series that will be repeatedly substituted later.
function SubCodePowers(SubCode,m);
	
	SubCodePow := [];
	for i in SubCode do
		SubCodePowi := [];
		for pair in i do
			e1 := pair[1];
			e2 := pair[2];
			e2pow := e2^0;
			e2vec := [e2pow];
			for i := 1 to m-1 do
				e2pow := e2pow*e2;
				Append(~e2vec,e2pow);
			end for;
			Append(~SubCodePowi,[* e1 , e2vec *]);
		end for;
		Append(~SubCodePow,SubCodePowi);
	end for;
	
	return SubCodePow;		

end function;

// Evaluate the power series f at some value, using precomputed powers.
function FastEvaluate(f,SubCodePowjk,m)

	fEval := Parent(f)!0;
	for i := 0 to m-1 do
		fEval := fEval + Coefficient(f,i)*SubCodePowjk[i+1];
	end for;

	return fEval;

end function;

// Input:
// p     -     A genus 0 prime, i.e. 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 41, 47, 59, or 71
// F     -     A binary quadratic form for which p is inert.
// m     -     A positive integer (the p-adic precision)
//
// Output:
// Fs    -     A list of power series, whose entries are
//                   a) Fs[1] = The product of factors (z - w1)/(z - w2) where w1,w2 run over SL_2(Z)-translates of F separated by 0.
//                      z = Standard variable on P1
//                   b) Fs[t] (for t>1) = A (p+1)-tuple of power series whose j-th entry is:
//                      Fs[t][j] = The product of factors (z - w1)/(z - w2) where w1,w2 run over SL_2(Z)-translates of F(p^t) separated by 0,
//                                 whose roots are in the residue disk of j. Expressed in terms of the variable
//                      x = 1/(z-j)    (if 0 <= j <= p-1)
//                      x = z          (if j = p+1, residue disk of infinity)

function rigidperiodfunction(F,p,m,K,sqrtDF,R,Kp,sqrtDFp,Rp : prt := false, nrm := false)
	
	// Power series of level 0 and 1.
	if prt eq true then
		print "Computing series in levels 0 and 1.";
	end if;
	Fs, Gs := Level01Series(F,p,m,K,sqrtDF,R,Kp,sqrtDFp,Rp : nrm := nrm);
	
	// Setting up the recursion.
	// STEP 1: Find SL_2(Z) matrices connecting -j/p -> oo to 0 -> oo.
	Gammas := []; 
	for j := 0 to p-1 do
		Append(~Gammas,GammasUnimodular(-j/p));
	end for;

	
	// STEP 2: List all the substitutions needed. 
	SubCode := []; 
	x := Rp.1;
	// Finite factors first. 
	for j := 0 to p-1 do
		jCode := [];
		for g in Gammas[j+1] do
			a := g[1,1];
			b := g[1,2];
			c := g[2,1];
			d := g[2,2];
			for l := 0 to p-1 do
				gl := MatrixActionModp(g,l,p);
				if gl eq p then
					f := (a + b*p*x)*(c + d*p*x)^(-1);
				else
					f := (c + d*p*x)*(a + b*p*x - gl*(c + d*p*x))^(-1);
				end if;
				Append(~jCode,[*gl,f*]);
			end for;
		end for;
		Append(~SubCode,jCode);
	end for;
	// Infinite factor.
	infCode := [];
	for l := 1 to p-1 do
		Append(~infCode,[*l,(p*x-l)^(-1)*]);
	end for;
	Append(~infCode,[*p,p*x*]);
	Append(~SubCode,infCode);	
	if prt eq true then
		print "Computing substitutions.";
	end if;
	SubCodePow := SubCodePowers(SubCode,m);
			
	// STEP 3: Iterate Up.
	for i := 2 to m-1 do
		if prt eq true then
			print "Doing level ", i, " now: " cat ("=")^(20-#IntegerToString(i));
		end if;
		Fi := [];
		Gi := [];
		FsPrev := Fs[#Fs];
		GsPrev := Gs[#Gs];
		for j := 0 to p do
			Fij := Rp!1;
			Gij := Rp!1;
			for pairs in SubCodePow[j+1] do	
				Fij := Fij*FastEvaluate(GsPrev[pairs[1]+1],pairs[2],m);
				Gij := Gij*FastEvaluate(FsPrev[pairs[1]+1],pairs[2],m);
			end for;
			Append(~Fi,Fij/Coefficient(Fij,0));
			Append(~Gi,Gij/Coefficient(Gij,0));
		end for;
		Append(~Fs,Fi);
		Append(~Gs,Gi);
	end for;
	
	FsMult := MultiplySeries(Fs,p);
	GsMult := MultiplySeries(Gs,p);
	
	return [* FsMult, GsMult, sqrtDF, sqrtDFp, F, p, m, Fs, Gs *];

end function;


// Construct data for the p-adic rigid meromorphic cocycle attached to F, to precision m;
function RigidMeromorphicCocycle(F,p,m : prt := true)

	DF := Discriminant(F);
	assert KroneckerSymbol(DF,p) eq -1; // Prime must be inert in order of F!
	// Local universes
	Kp<s>   := UnramifiedExtension(pAdicRing(p,m),2);
	Pol<t>  := PolynomialRing(Kp);
	sqrtDFp := Roots(t^2 - DF,Kp)[1][1];
	Rp<x>   := PowerSeriesRing(Kp,m);
//  Qp          := pAdicRing(p,m);
// 	R<tp>       := PolynomialRing(Qp);
// 	Kp<sqrtDFp> := ext<Qp | tp^2 - DF>;

	// Global universes
	QQ        := Rationals();
	RR<t>     := PolynomialRing(QQ);
	K<sqrtDF> := ext<QQ | t^2 - DF>;
	R<z>      := PolynomialRing(K);
	
	return rigidperiodfunction(F,p,m,K,sqrtDF,R,Kp,sqrtDFp,Rp : prt := prt);

end function;

// Abbreviation for Henri.
function rmc(F,p,m : prt := true);

	return RigidMeromorphicCocycle(F,p,m : prt := true);

end function;


// Construct data for the p-adic analytic meromorphic cocycle attached to F, to precision m;
function RigidAnalyticCocycle(F,p,m : prt := true)

	DF := Discriminant(F);
	assert KroneckerSymbol(DF,p) eq 1; // Prime must be split in order of F!
	// Local universes
	Zp      := IntegerRing(p^m);
	Pol<t>  := PolynomialRing(Integers());
  	sqrtDFp := Zp!(Integers()!Roots(t^2-DF,pAdicRing(p,m+10))[1][1]);
  	Rp<x>   := PowerSeriesRing(Zp,m);
	// Global universes
	QQ        := Rationals();
	RR<t>     := PolynomialRing(QQ);
	K<sqrtDF> := ext<QQ | t^2 - DF>;
	R<z>      := PolynomialRing(K);
	
	return rigidperiodfunction(F,p,m,K,sqrtDF,R,Zp,sqrtDFp,Rp : prt := prt);

end function;


// Construct data for the "norm" p-adic rigid meromorphic cocycle attached to F, to precision m;
function NormCocycle(F,p,m : prt := true)

	DF := Discriminant(F);
	// Local universes
	Zp      := IntegerRing(p^m);
  	sqrtDFp := Zp!0; // Not actually needed for the construction of the cocycle!
  	Rp<x>   := PowerSeriesRing(Zp,m);
	// Global universes
	QQ        := Rationals();
	RR<t>     := PolynomialRing(QQ);
	K<sqrtDF> := ext<QQ | t^2 - DF>; 
	R<z>      := PolynomialRing(K);
	
	return rigidperiodfunction(F,p,m,K,sqrtDF,R,Zp,sqrtDFp,Rp : prt := prt, nrm := true);

end function;

// MAIN FUNCTION
// -----------------------------------------------------------------------
// RigidCocycle(F,p,m : prt := true, nrm := false)
//
// Input: F   -   Indefinite binary quadratic form
//        p   -   prime
//        m   -   precision
//
// Output: Data for p-adic rigid cocycle attached to F, to precision m. 
// -----------------------------------------------------------------------
function RigidCocycle(F,p,m : prt := true, nrm := false)

	assert IsPrime(p);
	DF := Discriminant(F);
	
	if nrm then
		return NormCocycle(F,p,m : prt := prt);
	end if;

	// Split cocycles
	if KroneckerSymbol(DF,p) eq 1 then
		if p eq 2 then
			print "Split cocycles for p=2 are currently not supported.";
			return 0;
		else
			return RigidAnalyticCocycle(F,p,m : prt := prt);
		end if;
	end if;
	
	// Inert cocycles
	if KroneckerSymbol(DF,p) eq -1 then
		return RigidMeromorphicCocycle(F,p,m : prt := prt);
	end if;	
	
	// Ramified cocycles
	if KroneckerSymbol(DF,p) eq 0 and nrm eq false then
		print "Ramified primes currently not supported, *unless* you add the option 'nrm := true'. ";
		return 0;
	end if;
	
end function;



// ============================================================
// ===                7) Writing functions                  ===
// ============================================================

// These functions write the data structure of a computed theta series to a readable file.

// Convert number in quadratic extension K to string.
function PrintCoeff(a)

	ZZ := Integers();
	a1num := ZZ!Numerator(a[1]);
	a1den := ZZ!Denominator(a[1]);
	a2num := ZZ!Numerator(a[2]);
	a2den := ZZ!Denominator(a[2]);
	
	return "(" cat IntegerToString(a2num) cat "/" cat IntegerToString(a2den) cat "*sqrtDF + " cat IntegerToString(a1num) cat "/" cat IntegerToString(a1den) cat ")";

end function;

// Convert number in quadratic extension Kp to string.
function PrintCoeffp(a)

	ZZ := Integers();
	if Type(Parent(a)) eq RngIntRes then
	
		return IntegerToString(ZZ!a);
		
	else
		a1 := ZZ!Coefficient(a,1);
		a2 := ZZ!Coefficient(a,2);
		
		return "(" cat IntegerToString(a2) cat "*s + " cat IntegerToString(a1) cat ")";
		
	end if;

end function;


// Write the computed cocycle to a readable file.
function WriteCocycle(Phi,str)

	read store,"Are you sure you want to save this cocycle? This will overwrite any previously computed cocycle with the same name and prime. ('y' or 'n')";

	if store eq "y" then
		// Parent structures.
		Kp := Parent(Phi[4]);
		F  := Phi[5];
		DF := Discriminant(F);
		p  := Phi[6];
		m  := Phi[7];
			
		filename:="./J" cat str cat "p" cat IntegerToString(p) cat ".m";
		// Overwrites any previously stored data for the same parameters. 
		Write(filename,"load \"rmmc.m\";" : Overwrite := true);
		Write(filename,"");
		
		Write(filename, "DF        := " cat IntegerToString(DF) cat";");
		Write(filename, "QQ        := QuadraticForms(" cat IntegerToString(DF) cat ");");
		Write(filename, "Q         := Rationals();");
		Write(filename, "RR<t>     := PolynomialRing(Q);");
		Write(filename, "K<sqrtDF> := ext<Q | t^2 - DF>;");
		Write(filename, "R<z>      := PolynomialRing(K);");
		Write(filename, "");

		Write(filename, "p := " cat IntegerToString(p) cat ";");
		Write(filename, "m := " cat IntegerToString(m) cat ";");
		if Type(Kp) eq RngIntRes then
			Write(filename, "Kp := IntegerRing(p^m);");
		else
			Write(filename, "Kp<s> := UnramifiedExtension(pAdicRing(p,m),2);");
		end if;
		Write(filename, "Rp<x> := PowerSeriesRing(Kp,m);");
		Write(filename, "");
		
		Write(filename, "J" cat str cat "p" cat IntegerToString(p) cat " := [* [* ");
		
		// ** EVEN LEVELS
		// Level 0 contribution.
		Num := Numerator(Phi[1][1]);
		Den := Denominator(Phi[1][1]);
		f := "(" cat PrintCoeff(Coefficient(Num,0));
		g := PrintCoeff(Coefficient(Den,0));
		for i := 1 to Degree(Num) do
			f := f cat "+" cat PrintCoeff(Coefficient(Num,i)) cat "*z^" cat IntegerToString(i);
			g := g cat "+" cat PrintCoeff(Coefficient(Den,i)) cat "*z^" cat IntegerToString(i);
		end for;
		Write(filename, f cat ")/(" cat g cat "), [");
		// Power series on disks.
		for j := 1 to p+1 do
			Fj := PrintCoeffp(Coefficient(Phi[1][2][j],0));
			for i := 1 to m-1 do
				Fj := Fj cat "+" cat PrintCoeffp(Coefficient(Phi[1][2][j],i)) cat "*x^" cat IntegerToString(i);
			end for;
			Write(filename, Fj);
			if j ne p+1 then
				Write(filename, ",");
			end if;
		end for;	
							
		// ** ODD LEVELS
		// Level 0 contribution.
		Num := Numerator(Phi[2][1]);
		Den := Denominator(Phi[2][1]);
		f := "(" cat PrintCoeff(Coefficient(Num,0));
		g := PrintCoeff(Coefficient(Den,0));
		for i := 1 to Degree(Num) do
			f := f cat "+" cat PrintCoeff(Coefficient(Num,i)) cat "*z^" cat IntegerToString(i);
			g := g cat "+" cat PrintCoeff(Coefficient(Den,i)) cat "*z^" cat IntegerToString(i);
		end for;
		Write(filename, "] *], [* " cat f cat ")/(" cat g cat "), [");		
		// Power series on disks.
		for j := 1 to p+1 do
			Fj := PrintCoeffp(Coefficient(Phi[2][2][j],0));
			for i := 1 to m-1 do
				Fj := Fj cat "+" cat PrintCoeffp(Coefficient(Phi[2][2][j],i)) cat "*x^" cat IntegerToString(i);
			end for;
			Write(filename, Fj);
			if j ne p+1 then
				Write(filename, ",");
			end if;
		end for;	
					
		// sqrtDF, sqrtDFp, F, p, m 
		Write(filename, "] *], sqrtDF, " cat "Kp!" cat PrintCoeffp(Phi[4]));
		Write(filename, ", QQ!<" cat IntegerToString(F[1]) cat "," cat IntegerToString(F[2]) cat "," cat IntegerToString(F[3]) cat ">");
		Write(filename, ", " cat IntegerToString(p) cat ", " cat IntegerToString(m) cat " *]");		
		Write(filename,";");
		
	end if;
	
	return "Done. ";

end function;


// ============================================================
// ===                      8) Tests                        ===
// ============================================================

function StandardTest(F,p,m : nrm := false)

	Phi := RigidCocycle(F,p,m : prt := false, nrm := nrm);
	DF  := Discriminant(Phi[5]); 
	Stest1  := SubstitutionProduct(Phi[1],[IMat,SMat]);
	Stest2  := SubstitutionProduct(Phi[2],[IMat,SMat]);
	STtest1 := SubstitutionProduct(Phi[1],[IMat,STMat,STMat^2]);
	STtest2 := SubstitutionProduct(Phi[2],[IMat,STMat,STMat^2]);
	K  := Parent(Phi[3]);
	Kp := Parent(Phi[4]);
	
// 	print Stest1, Stest2, STtest1, STtest2;
		
	for i := 1 to m-1 do
		a := K!Stest1[1];  // Will throw error if level 0 is not trivial.
		a := K!Stest2[1];  // Will throw error if level 0 is not trivial.
		a := K!STtest1[1]; // Will throw error if level 0 is not trivial.
		a := K!STtest2[1]; // Will throw error if level 0 is not trivial.
		for j := 1 to p+1 do
			if p ne 2 or KroneckerSymbol(DF,p) eq 1 then
				assert Coefficient(Stest1[2][j],i)  eq 0;
				assert Coefficient(Stest2[2][j],i)  eq 0;
				assert Coefficient(STtest1[2][j],i) eq 0;
				assert Coefficient(STtest2[2][j],i) eq 0;
			else
				assert Valuation(Coefficient(Stest1[2][j],i))  ge m-2;
				assert Valuation(Coefficient(Stest2[2][j],i))  ge m-2;
				assert Valuation(Coefficient(STtest1[2][j],i)) ge m-2;
				assert Valuation(Coefficient(STtest2[2][j],i)) ge m-2;
			end if;				
		end for;
	end for;
	
	return Phi;
	
end function;



function TestSplitDiscriminants(p,I,m)

	Ds := SplitDiscriminants(p,I);
	for D in Ds do
		print "Testing D =", D;
		Red := ReducedOrbits(QuadraticForms(D));
		for orb in Red do
			F := orb[1];
			a := StandardTest(F,p,m);
		end for;
	end for;
	
	return "";

end function;
