// =============================================================
// =============================================================
//
// This file calls the cocycles in H^1(Gamma,M*/Cp*) defined by the rigid meromorphic period functions
//
//       J_F(S)(z) = prod [z-w_G]^sgn(G)
//
// and investigates their lifting obstruction to H^1(Gamma,M*)


// ============================================================
// ===                 Table of contents                    ===
// ============================================================

//    (1) The involutions winf and wp
//    (2) Hecke operators away from p

// ====================================================================
// ===            1) The involutions winf, wp, and w_rec            ===
// ====================================================================


// Compute (1+wp)Phi, for a cocycle Phi.
function wpPlus(Phi)
	
	pPlus := DivideList(Phi[1],Phi[2]);
	return [* pPlus, PowerList(pPlus,-1), Phi[3], Phi[4], Phi[5], Phi[6], Phi[7] *];

end function;


// Compute (1-wp)Phi, for a cocycle Phi.
function wpMinus(Phi)

	pMin := MultiplyList(Phi[1],Phi[2]);
	return [* pMin, pMin, Phi[3], Phi[4], Phi[5], Phi[6], Phi[7] *];

end function;


// Compute (1+w_inf)Phi, for a cocycle Phi.
function winfPlus(Phi)

	p := Phi[6];
	winf := Matrix(2,2,[[-1, 0],[0,1]]);  // Involution.
	PhiPlus1 := SubstitutionProduct(Phi[1],[winf]);
	PhiPlus2 := SubstitutionProduct(Phi[2],[winf]);
	iPlus1   := DivideList(Phi[1],PhiPlus1);
	iPlus2   := DivideList(Phi[2],PhiPlus2);
	return [* iPlus1, iPlus2 , Phi[3], Phi[4], Phi[5], Phi[6], Phi[7] *];
		
end function;


// Compute (1-w_inf)Phi, for a cocycle Phi.
function winfMinus(Phi)

	p := Phi[6];
	winf := Matrix(2,2,[[-1, 0],[0,1]]);  // Involution.
	Min1 := SubstitutionProduct(Phi[1],[IMat,winf]);
	Min2 := SubstitutionProduct(Phi[2],[IMat,winf]);
	return [* Min1, Min2, Phi[3], Phi[4], Phi[5], Phi[6], Phi[7] *];
	
end function;


// The following allows us to compute the cocycle which is symmetrised with 
// respect to the 'involution' of reciprocation. 

// Conjugates the power series over Q_p^2.
function iConj(Fs)

	z  := Parent(Fs[1]).1;
	x  := Parent(Fs[2][1]).1;
	K  := Parent(Coefficient(Numerator(Fs[1]),0));
	Kp := Parent(Coefficient(Fs[2][1],0));
	s  := K.1;
	sp := Kp.1;
	sb := Norm(s)/s;
	spb:= Norm(sp)/sp;
	
	// Level 0
	Num := Numerator(Fs[1]);
	Numb := 0*z;
	for i := 0 to Degree(Num) do
		ai := Coefficient(Num,i);
		Numb := Numb + (ai[2]*sb + ai[1])*z^i;
	end for;
	Den := Denominator(Fs[1]);
	Denb := 0*z;
	for i := 0 to Degree(Den) do
		ai := Coefficient(Den,i);
		Denb := Denb + (ai[2]*sb + ai[1])*z^i;
	end for;
	
	// Mittag-Leffler
	ML := [];
	for f in Fs[2] do
		fb := 0*x;
		for i := 0 to Degree(f) do
			ai := Coefficient(f,i);
			fb := fb + (Coefficient(ai,2)*spb + Coefficient(ai,1))*x^i;
		end for;
		Append(~ML,fb);
	end for;		
	
	return [* Numb/Denb, ML *];
	
end function;


// Compute (1+w_rec)Phi, for a cocycle Phi.
function wrecPlus(Phi)

	iPlus1 := DivideList(Phi[1],iConj(Phi[1]));
	iPlus2 := DivideList(Phi[2],iConj(Phi[2]));
	return [* iPlus1, iPlus2 , Phi[3], Phi[4], Phi[5], Phi[6], Phi[7] *];
		
end function;


// Compute (1-w_rec)Phi, for a cocycle Phi.
function wrecMinus(Phi)

	iMin1 := MultiplyList(Phi[1],iConj(Phi[1]));
	iMin2 := MultiplyList(Phi[2],iConj(Phi[2]));
	return [* iMin1, iMin2, Phi[3], Phi[4], Phi[5], Phi[6], Phi[7] *];
	
end function;


// ====================================================================
// ===               2) Hecke operators away from p                 ===
// ====================================================================

// Return the cocycle T_l(Phi), l not p.
function HeckeCocycle(Phi,l) 

	assert IsPrime(l);
	
	// STEP 1: Find SL_2(Z) matrices connecting -j/l -> oo to 0 -> oo.
	Gammas := [Matrix(2,2,[[l, 0],[0,1]])]; 
	for j := 0 to l-1 do
		for g in GammasUnimodular(-j/l) do
			Append(~Gammas,g*Matrix(2,2,[[1, -j],[0,l]]));
		end for;
	end for;
	TlPhi := SubstitutionCocycle(Phi,Gammas);
	
	return TlPhi;	
	
end function;

// Hecke projector under very specific operator:
//        T  = (T_l1 - a1)*(T_l2 - a2)*....*(Tln - an)
// This is encoded in L = [ [l1,a1], [l2,a2], ... ]
function HeckeProjection(Phi,L)

	if Type(L[1]) eq RngIntElt then
		L := [L];
	end if;
	
	TPhi := Phi;
	for M in L do
		TM1	 := HeckeCocycle(TPhi,M[1]);
		TM2  := CocyclePower(TPhi,M[2]);
		TPhi := CocycleDivision(TM1,TM2);
	end for;

	return NormaliseCocycle(TPhi); // Normalise the cocycle so that Obstructions(TPhi) is meaningful.

end function;

// Compute the lifting obstruction (Tl - (l+1))kappa{0,oo} 
// for all primes l in the list 'ls'.

// function CuspidalLiftingObstructions(Phi, ls)
// 
// 	p := Phi[6];
// 	for l in ls do
// 		assert IsPrime(l);
// 		assert p ne l;
// 		TlPhi := HeckeProjection(Phi,[[l,l+1]]);
// 		CS, CST := Obstructions(TlPhi);
// 		print "--------------------";
// 		print " Prime l =", l;
// 		print " Lifting obstruction: ";
// 		P1 := algdep(CS[1]*CS[2],4); 
// 		print P1;
// 		for i := 0 to Degree(P1) do
// 			print "Coeff", i, "is", Factorisation(Coefficient(P1,i));
// 		end for;
// 		
// 		print " Also, the lifting obstruction at elliptic points satisfies: ", algdep(CS[1]^3/CST[1]^2,2);
// 	end for;
// 	
// 	return "Done. ";
// 
// end function;


// // Compute the primitive "l-derived" forms of F, and return their reduction orbits.
// function TlQuadraticForms(F,l)
// 
// 	DF := Discriminant(F);
// 	a := F[1];
// 	b := F[2];
// 	c := F[3];
// 	L := { };
// 	Lcount := 0;
// 	for j := 0 to l-1 do
// 		QQjl := QuadraticForms(DF*l^2)!<a*l^2,l*(b+2*j*a),a*j^2+b*j+c>;
// 		if IsPrimitiveForm(QQjl) then
// 			L := L join {ReductionOrbit(QQjl)};
// 			Lcount := Lcount+1;
// 		end if;
// 	end for;
// 	QQinfl := QuadraticForms(DF*l^2)!<a,b*l,c*l^2>;
// 	if IsPrimitiveForm(QQinfl) then
// 		L := L join {ReductionOrbit(QQinfl)};
// 		Lcount := Lcount+1;
// 	end if;
// 	
// 	return L,Lcount;
// 
// end function;
