// =============================================================
// =============================================================
//
// This file uses the cocycles in H^1(Gamma,M*/Cp*) defined by the rigid period functions
//
//       J_F(S)(z) = prod [z-w_G]^sgn(G) 
//
// implemented in "rmmc_Cocycle.m", and computes their values.
// This encompasses three particular situations:
//
//   a) RM values
//          Well-defined up to small ambiguity contained in eps^Z
//          These should lie in appropriate ring class fields, and
//          satisfy a Shimura reciprocity law.
//          See: (1) - (2)
//
//   b) Values at infinity ("Cuspidal values")
//          These are, up to a small power, p-units in the ring class field of F
//          See: (4)
//
//   c) Values of the parabolic lifting obstruction kappa_tau[0,infty]
//          Its components along the classes of elliptic curves E 
//          are well-defined up to powers of the Tate parameter q.
//          These are, up to an explicit factor, Stark-Heegner points on E.
//          See: (5)


// ============================================================
// ===                 Table of contents                    ===
// ============================================================

//    (1) RM Values
//    (2) Shimura reciprocity
//    (3) Obstructions
//    (4) Cuspidal Values
//    (5) Stark-Heegner points
//    (6) Slope 1 invariants
//    (7) Tests

// =============================================================
// ===                    1) RM Values                       ===
// =============================================================

// Function to evaluate power series, which is necessary only in case where p is split,
// since the power series is then defined over IntegerRing(p^m). 

function evaluate_series(f,tau,spl);

	if spl eq false then
		return Evaluate(f,tau);
	else
		R := Parent(f);
		K := Parent(tau);
		m := Precision(R);
		Eval := K!0;
		for i := 0 to m-1 do
			Eval := Eval + K!(Integers()!Coefficient(f,i))*tau^i;
		end for;
	
		return Eval;
	end if;
	
end function;

// Function to evaluate the rational period function.

function value_period_function(Fs,wG,wGp,p,spl)
 		
 	Int0 := 1; // Indefinite integral, level 0.
 	Intp := 1; // Indefinite integral, level >0.
 	
	// Evaluate series level 0.  
	f0gl := Evaluate(Fs[1],wG-1)/Evaluate(Fs[1],1-1/wG);
	Int0 := Int0*f0gl;

	// Evaluate series at all higher levels.
	for l := 0 to p-1 do  // Finite factors
		Fsl := Fs[2][l+1];
		fl  := evaluate_series(Fsl,1/(wGp-1-l),spl)/evaluate_series(Fsl,1/(1-1/wGp-l),spl);
		Intp:= Intp*fl;
	end for;
	// Infinite factor
	Fsinf := Fs[2][p+1];
	finf  := evaluate_series(Fsinf,wGp-1,spl)/evaluate_series(Fsinf,1-1/wGp,spl);
	Intp:= Intp*finf;
	
	return Int0, Intp;

end function;

// Calculates the evaluation of the cocycle Fs at a form G.
function value_cocycle(Fs,G,p,sqrtDG,sqrtDFp,sqrtDGp,Kp,spl,boo)

	Red  := ReductionOrbit(G);
	Int0 := 1;
	Intp := 1;
	for Gi in Red do
		wG  := (-Gi[2]+sqrtDG )/(2*Gi[1]);
		wGp := (-Gi[2]+sqrtDGp)/(2*Gi[1]);
		Int0i, Intpi := value_period_function(Fs,wG,wGp,p,spl);
		Int0 := Int0*Int0i;
		Intp := Intp*Intpi;
 	end for;
	
	if boo eq true then
		Int0p := Kp!Int0[1][1] + Kp!Int0[1][2]*sqrtDFp + Kp!Int0[2][1]*sqrtDGp + Kp!Int0[2][2]*sqrtDFp*sqrtDGp;
	else
		Int0p := Kp!Int0[1] + Kp!Int0[2]*sqrtDFp;
	end if;
	
  	return Int0p*Intp;
  	
end function;


// Calculates the evaluation of the cocycle Phi at a form G.
function RMValue(Phi,G : alg := false)

	// Data from cocycle.
	Fs     := Phi[1];
	FsFrob := Phi[2];
	sqrtDF := Phi[3];
	DF := Discriminant(Phi[5]);
	p  := Phi[6];
	m  := Phi[7];
	K  := Parent(sqrtDF);
	Kp := Parent(Phi[4]);
	
	DG := Discriminant(G);
	assert KroneckerSymbol(DG,p) eq -1; // The roots of G must lie in the standard affinoid!

	// Local roots
	if Type(Kp) eq RngIntRes then
		Kp<s>   := UnramifiedExtension(pAdicField(p,m),2);
	   	sqrtDFp := Kp!Integers()!Phi[4];
   		spl := true;
   	else
   		sqrtDFp := Phi[4];
		Kp  := FieldOfFractions(Parent(Coefficient(Fs[2][1],0)));
   		spl := false;
	end if;	
	Pol<tp> := PolynomialRing(Kp);
   	sqrtDGp := Roots(tp^2 - DG,Kp)[1][1];
	
	// Global roots
 	QQ    := Rationals();
 	RR<t> := PolynomialRing(K);
 	if IsIrreducible(t^2-DG) then
 		KG<sqrtDG> := ext<K | t^2 - DG>;
 		boo        := true;
 	else
 		sqrtDG := Roots(t^2 - DG,K)[1][1];
 		boo := false;
 	end if;	
 	
 	a := value_cocycle(Fs,G,p,sqrtDG,sqrtDFp,sqrtDGp,Kp,spl,boo);	
 	
 	if alg eq true then
 	 	// Roots and fundamental units 
		zeta  := Roots(CyclotomicPolynomial(p^2-1),Kp)[1][1];
		fepsF := MinimalPolynomial(FundamentalUnit(QuadraticField(DF)));
		eps   := Roots(fepsF,Kp)[1][1];
		RedF := ReducedOrbits(QuadraticForms(DF));
		RedG := ReducedOrbits(QuadraticForms(DG));
		P1, A := ThoroughAlgdepEpsilon(a,4*#RedF*#RedG,DF);
		if Degree(P1) eq 0 then
        	print "Minimal polynomial:  Not recognised...";
    	else
        	print "Minimal polynomial (over Q)  : ", P1;
 		end if;
 	end if;
 	
 	return a;

end function;


// =============================================================
// ===                2) Shimura reciprocity                 ===
// =============================================================



// This function computes the elusive quantities Jp(F,G), for fixed F and all G of discriminant DG,
// and tests "Shimura reciprocity" by checking whether they satisfy a small common minimal polynomial. 

function shimura_reciprocity(Phi,
        DF,RedF,sqrtDF,sqrtDFp,// Invariants for F
        DG,RedG,sqrtDG,sqrtDGp,// Invariants for G
        K,Kp,zeta,eps,depth,// Universes, ambiguities, and thoroughness of algdep
        spl,boo,sym);// Boolean expressions for invariant type
        
	// Attempt to recognise ONE invariant.
	Fs      := Phi[1];
	p       := Phi[6];
	m       := Phi[7];
	PolZ<x> := PolynomialRing(Integers());
	P1      := 1*x^0;
	Inv     := [];
	i       := 0;
	while Degree(P1) eq 0 and i lt #RedG do
		i := i + 1;
		G := RedG[i][1];
		a := value_cocycle(Fs,G,p,sqrtDG,sqrtDFp,sqrtDGp,Kp,spl,boo);
		if sym then
			P1, A := ThoroughAlgdepZeta(a,#RedF*#RedG);
		else
			P1, A := ThoroughAlgdepEpsilon(a,4*#RedF*#RedG,DF : depth := depth);
		end if;
	end while;
		
	if Degree(P1) eq 0 then
        print "Minimal polynomial:  Not recognised...";
    else
        print "Minimal polynomial (over Q)  : ", P1; 
		Shimura_test := true;
		for GG in RedG do
			G := GG[1];
			a := value_cocycle(Fs,G,p,sqrtDG,sqrtDFp,sqrtDGp,Kp,spl,boo);
			if sym then
				Currl := 0;
				for l := 0 to p^2-2 do
					A := a*zeta^l;
					if Valuation(Evaluate(P1,A)) gt Floor(5*m/7) then
						Currl := l;
					end if;
				end for;
				A := a*zeta^Currl;
			else
				Currk := -depth;
				Currl := 0;
				done  := false;
				k     := -depth;
				while k lt depth and done eq false do
					k := k+1;
					for l := 0 to p^2-2 do
						A := a*zeta^l*eps^k;
						if Valuation(Evaluate(P1,A)) gt Floor(5*m/7) then
							Currk := k;
							Currl := l;
							done  := true;
						end if;
					end for;
				end while;
				A := a*zeta^Currl*eps^Currk;
			end if;
					
			if Valuation(Evaluate(P1,A)) lt Floor(5*m/7) then
				Shimura_test := false;
				print "Shimura failure for G =", G;
			end if;
			Append(~Inv,A);		
		end for;
		
		if Shimura_test eq false and IsCoprime(FundamentalDiscriminant(DG),DF) eq false then
			ValuationVectors(P1);
			print "Invariants are not Galois conjugate, but this *could* be caused by the fact that the discriminants are not coprime. ";
			print "Be critical of the output. ";
		end if;
		
		if Shimura_test eq false and IsCoprime(FundamentalDiscriminant(DG),DF) then
			print "";
			print "WARNING: Shimura reciprocity seems to fail. ";
			print "Polynomial over Q might not have been recognised correctly! ";
			print "Either try running the code again, or compute to higher precision. ";
		end if;
	
		if Shimura_test eq true then
			ValuationVectors(P1);
			print "Shimura reciprocity is satisfied for these invariants.";
		end if;
			
	end if;
	
	return Inv;
	
end function;


// Loop Jp over all suitable small discriminants.
function RMValues(Phi,I: sym := false, depth := 120)

	// Initialise all invariants. 
	sqrtDF := Phi[3];
	Kp   := Parent(Phi[4]);
	DF   := Discriminant(Phi[5]);
	p    := Phi[6];
	m    := Phi[7];
	RedF := ReducedOrbits(QuadraticForms(DF));
	
	// Polynomial rings, appropriately coerced for meromorphic vs analytic.
	if Type(Kp) eq RngIntRes then
		Kp<s>   := UnramifiedExtension(pAdicField(p,m),2);
	   	sqrtDFp := Kp!Integers()!Phi[4]; 
   		spl := true;
   	else
   		sqrtDFp := Phi[4];
		Kp  := FieldOfFractions(Parent(Coefficient(Phi[1][2][1],0)));
   		spl := false;
	end if;	
	Polp<tp> := PolynomialRing(Kp);
	K        := Parent(sqrtDF);
 	Pol<t>   := PolynomialRing(K);
 	
 	// Roots and fundamental units 
	zeta  := Roots(CyclotomicPolynomial(p^2-1),Kp)[1][1];
	eps   := 1;
	if sym eq false then
  		fepsF := MinimalPolynomial(FundamentalUnit(QuadraticField(DF)));
		eps   := Roots(fepsF,Kp)[1][1];
	end if;

	// Make input into sequence of discriminants, pick out inert ones. 
	if Type(I) eq RngIntElt then
		I := [I];
	end if;
	E := InertDiscriminants(p,I : DF := DF);
	assert #E gt 0; // p needs to be an inert prime for DG!
    
   	Invs := [* *];
    for DG in E do
    	RedG := ReducedOrbits(QuadraticForms(DG));
		hG   := #RedG;
   		sqrtDGp := Roots(tp^2 - DG,Kp)[1][1]; // Local root
   		if IsIrreducible(t^2-DG) then
 			KG<sqrtDG> := ext<K | t^2 - DG>;  // Global root
 			boo        := true;
 		else
 			sqrtDG := Roots(t^2 - DG,K)[1][1]; // Global root
 			boo := false;
 		end if;	

		print "=========================================================";
		print " Algebraic invariants Jp(tau1,tau2) ";
		print "      tau1 = root of ", Phi[5][1]*t^2+Phi[5][2]*t+Phi[5][3];
		print "      tau2 = runs through roots of the", #ReducedOrbits(QuadraticForms(DG)), "classes of discriminant", DG;
		print "(Precision:", p, "^", m, ")";
		print " "; 
        InvPhi := shimura_reciprocity(Phi,
        DF,RedF,sqrtDF,sqrtDFp,// Invariants for F
        DG,RedG,sqrtDG,sqrtDGp,// Invariants for G
        K,Kp,zeta,eps,depth,// Universes, ambiguities, and thoroughness of algdep
        spl,boo,sym);// Boolean expressions for invariant type
        Append(~Invs, InvPhi);
		print " ";
    end for;
   
// 	return Invs;
	return " ";

end function;


// =============================================================
// ===                    3) Obstructions                    ===
// =============================================================

// Computes the products of the constant terms in a list L.
function constants_product(L,K,Kp,sqrtDFp,m)
	
	C0 := K!L[1]; // Level 0 constant, implicit test
	C  := Kp!C0[1]+Kp!C0[2]*sqrtDFp;
	// Higher levels
	for i := 1 to #L[2] do
		C := C*Kp!Coefficient(L[2][i],0);
		for j := 1 to m-1 do
			assert Coefficient(L[2][i],j) eq 0;
		end for;
	end for;

	return C;

end function;


// Computes the obstructions for a rigid meromorphic cocycle Phi to be a modular symbol for S and ST
function Obstructions(Phi : tests := false)

	K  := Parent(Phi[3]);
 	Zp := Parent(Phi[4]);
	DF := Discriminant(Phi[5]);
 	p  := Phi[6];
 	m  := Phi[7];
 	
 	// Polynomial rings, appropriately coerced for meromorphic vs analytic.
	if Type(Zp) eq RngIntRes then
		Kp<s>   := UnramifiedExtension(pAdicField(p,m),2);
	   	sqrtDFp := Kp!Integers()!Phi[4];
   	else
		Kp<s>   := FieldOfFractions(Zp);
   		sqrtDFp := Phi[4];
	end if;	

	// Test cocycle relations for S, order 4 element
	Sub := [SubstitutionProduct(Phi[i],[IMat,SMat])  : i in [1,2]];	
	CS  := [constants_product(Sub[i],K,Kp,sqrtDFp,m) : i in [1,2]];
	print "S sanity test passed. ";
	// Test cocycle relations for ST, order 6 element
	Sub := [SubstitutionProduct(Phi[i],[IMat,STMat,STMat^2]) : i in [1,2]];	
	CST := [constants_product(Sub[i],K,Kp,sqrtDFp,m)         : i in [1,2]];
	print "ST sanity test passed. ";
	// Test cocycle relations for S directly:
	C0  := [Evaluate(Phi[i][1],0)           : i in [1,2]];
	CSd := [Kp!C0[i][1]+Kp!C0[i][2]*sqrtDFp : i in [1,2]];
	for j := 1 to p-1 do
		CSd[1] := CSd[1]*Kp!Evaluate(Phi[1][2][j+1],Zp!(-1/j));
		CSd[2] := CSd[2]*Kp!Evaluate(Phi[2][2][j+1],Zp!(-1/j));
	end for;
	
	assert CSd eq CS; // Test whether we got the same answer.
	print "Comparison test passed. ";
	
	// Sanity checks
	if tests then
		OO    := QuadraticOrder(QuadraticForms(DF));
		feps  := MinimalPolynomial(FundamentalUnit(OO));
		eps   := Roots(feps,Kp)[1][1];
		if Norm(eps) eq -1 then
			eps := eps^2;
		end if;
		k   := 6*(1-KroneckerSymbol(DF,p)*p);
		obs := (CS[1]^3/CST[1]^2)^(1-p^2);
		assert Valuation(obs-eps^k) ge m or Valuation(obs-eps^-k) ge m;
		k   := 6*(KroneckerSymbol(DF,p)-p);
		obs := (CS[2]^3/CST[2]^2)^(1-p^2);
		assert Valuation(obs-eps^k) ge m or Valuation(obs-eps^-k) ge m;
		print "Tests passed. ";
	end if;

 	return CS,CST;

end function;




// =============================================================
// ===                 4) Cuspidal values                    ===
// =============================================================

// Given a p-adic number 'a' which has some random generator 's' for the quadratic extension,
// display its coefficients with respect to a new generator 'sqrtDFp'.
function SwitchToNiceGenerator(a,p,m,sqrtDFp)

	Qp   := pAdicField(p,m);
	R<x> := PolynomialRing(Qp);
	Lp<sqDF> := ext<Qp | x^2 - Integers()!(sqrtDFp^2)>;
	
	b1 := Trace(a)/2;
	b2 := Qp!((a-b1)/sqrtDFp);
	
	return b1 + b2*sqDF;	

end function;


// Computes the "lifting obstruction" from the paper, which should be a Gross-Stark unit.
// In [DV2], it is proved that it is an algebraic unit. 
function CuspidalValue(Phi : alg := true) 

	K  := Parent(Phi[3]);
 	Zp := Parent(Phi[4]);
	DF := Discriminant(Phi[5]);
 	p  := Phi[6];
 	m  := Phi[7];
	RedF := ReducedOrbits(QuadraticForms(DF));
	 	
 	// Polynomial rings, appropriately coerced for meromorphic vs analytic.
	if Type(Zp) eq RngIntRes then
		Kp<s>   := UnramifiedExtension(pAdicField(p,m),2);
	   	sqrtDFp := Kp!Integers()!Phi[4];
   	else
		Kp<s>   := FieldOfFractions(Zp);
   		sqrtDFp := Phi[4];
	end if;	

	// Test cocycle relations for S directly:
	C0  := [Evaluate(Phi[i][1],0)           : i in [1,2]];
	GS  := [Kp!C0[i][1]+Kp!C0[i][2]*sqrtDFp : i in [1,2]];
	for j := 1 to p-1 do
		GS[1] := GS[1]*Kp!Evaluate(Phi[1][2][j+1],Zp!(-1/j));
		GS[2] := GS[2]*Kp!Evaluate(Phi[2][2][j+1],Zp!(-1/j));
	end for;
	
	if alg eq true then
		P1 := algdep(GS[1]*GS[2],#RedF);
		print "Minimal polynomial: ", P1;
// 		ValuationVectors(P1);
	end if;
	
	return GS[1]*GS[2];
	
end function;


// Computes the "lifting obstruction" from the paper, which should be a Gross-Stark unit.
// In the paper, it is proved that its *norm* is an algebraic unit. 
function CuspidalValues(DF,p,m : alg := false) 

	assert KroneckerSymbol(DF,p) eq -1; // Prime must be inert (for now!)
	// Local universes
	Qp          := pAdicRing(p,m);
	RRp<tp>     := PolynomialRing(Qp);
	Kp<sqrtDFp> := ext<Qp | tp^2 - DF>;
	Rp<x>       := PowerSeriesRing(Kp,m);
	// Global universes
	QQ        := Rationals();
	RR<t>     := PolynomialRing(QQ);
	K<sqrtDF> := ext<QQ | t^2 - DF>;
	R<z>      := PolynomialRing(K);
	
	RedF := ReducedOrbits(QuadraticForms(DF));
	Cocycles   := [* *];
	Invariants := [* *];
	
	for orb in RedF do
		F  := orb[1];
		print "Cuspidal value of quadratic form ", F;
		JF := rigidperiodfunction(F,p,m,K,sqrtDF,R,Kp,sqrtDFp,Rp);
		Append(~Cocycles,JF);
		// Test cocycle relations for S directly:
		C0  := [Evaluate(JF[i][1],0)            : i in [1,2]];
		GS  := [Kp!C0[i][1]+Kp!C0[i][2]*sqrtDFp : i in [1,2]];
		for j := 1 to p-1 do
			GS[1] := GS[1]*Kp!Evaluate(JF[1][2][j+1],-1/j);
			GS[2] := GS[2]*Kp!Evaluate(JF[2][2][j+1],-1/j);
		end for;
		if alg eq true then
			P1 := algdep(GS[1]*GS[2],#RedF);
			print "satisfies the polynomial: ", P1;
		end if;	
		Append(~Invariants,GS[1]*GS[2]);
	end for;
	
	return Invariants, Cocycles;
	
end function;


// ============================================================
// ===               5) Stark-Heegner points                ===
// ============================================================


function Fudge(E)
	
	Sha := Floor(ConjecturalRegulator(E)); 	
	Tam := 1;
	for no in TamagawaNumbers(E) do
		Tam := Tam*no;
	end for;
	fudge := Sha*Tam/Order(TorsionSubgroup(E))^2;
	
	r,Lr1 := AnalyticRank(E: Precision:=50);
	OmE   := RealPeriod(E);
	print fudge;
	print Lr1/OmE; // Check that this agrees with analytic calculation.
	
	return fudge;

end function;


// Computes the Stark--Heegner point on E of the cocycle J, as projected by the Hecke operator T.
// For now, there are no sanity checks, and the Hecke projector T has to be fed manually.
// TODO incorporate the Hecke projection.
function StarkHeegnerPoint(J,E)

	p := J[6];
	m := J[7];	
	// Elliptic curve invariants
	q  := qParameter(E,p,m);
	ETate    := EllipticCurve([1,0,0,TateA4(q,m),TateA6(q,m)]);
	Ep,f1,f2 := WeierstrassModel(ETate);
	
	// Value of the cocycle, test whether it is in q^Z.
	u  := CuspidalValue(J : alg := false);
	triv := false;
	for i := -m to m do
		if Valuation(u-q^i) ge m then 
			print "u: ", u, " q^i:", q^i;
			triv := true; // u maps to the origin of E.
		end if;
	end for;
	
	// Tate parametrisation for u.
	if triv eq false then 
		X := TateX(u,q);
		Y := TateY(u,q);
		Z := Parent(q)!1;
// 		print "Should be zero if point is on the Tate curve: ", Y^2 + X*Y - X^3 - TateA4(q,m)*X - TateA6(q,m);
		assert Valuation(Y^2 + X*Y - X^3 - TateA4(q,m)*X - TateA6(q,m)) ge m-1;
	else
		X := 0;
		Y := 1;
		Z := 0;
	end if;

	Pointp := Ep!f1(ETate![X,Y,Z]);
// 	print "Pointp is: ", Pointp;
	psi    := Isomorphism(Ep,ChangeRing(E,Parent(q)));
	Point  := psi(Pointp);
	
	DF := Discriminant(J[5]);
	RedF := ReducedOrbits(QuadraticForms(DF)); 
	print "X-coordinate satisfies : ", algdep(Point[1],2*#RedF);
	print "Y-coordinate satisfies : ", algdep(Point[2],2*#RedF);

	return Point;

end function;



function StarkHeegnerPoints(E,p,m,DF,T)
	
	assert KroneckerSymbol(DF,p) eq -1; // Prime must be inert (for now!)
	// Local universes
	Zp<s>   := UnramifiedExtension(pAdicRing(p,m),2);
  	Pol<t>  := PolynomialRing(Integers());
  	sqrtDFp := Roots(t^2 - DF,Zp)[1][1]; // CAREFUL!!! Sometimes Magma loses precision here.
	Rp<x>   := PowerSeriesRing(Zp,m);
	Kp      := FieldOfFractions(Zp);
	// Global universes
	QQ        := Rationals();
	RR<t>     := PolynomialRing(QQ);
	K<sqrtDF> := ext<QQ | t^2 - DF>;
	R<z>      := PolynomialRing(K);
	// Elliptic curve local information
	q        := Kp!qParameter(E,p,m);
	ETate    := EllipticCurve([1,0,0,Kp!TateA4(q,m),Kp!TateA6(q,m)]);
	ETateWei,f1,f2 := WeierstrassModel(ETate);
	Ep       := ChangeRing(E,Kp);
		
	RedF := ReducedOrbits(QuadraticForms(DF));
	Cocycles := [* *];
	Points   := [* *];
	
	for orb in RedF do

		// Compute Hecke projection to E-component.
		F   := orb[1];
		print "Now doing form", F;
		J   := rigidperiodfunction(F,p,m,K,sqrtDF,R,Zp,sqrtDFp,Rp);
		J_E := HeckeProjection(J,T);
		
		// Evaluation.
		u  := Kp!CuspidalValue(J_E : tests := false);
		triv := false;
		for i := -m to m do
			if Valuation(u-q^i) ge m then
				print "u: ", u, " q^i:", q^i;
				triv := true; // u maps to the origin of E.
			end if;
		end for;
		
		// Tate uniformisation.
		if triv eq false then // Tate parametrisation for u.
			X := TateX(u,q);
			Y := TateY(u,q);
			Z := 1;
			print "Should be zero if point is on the Tate curve: ", Y^2 + X*Y - X^3 - TateA4(q,m)*X - TateA6(q,m);
		else
			X := 0;
			Y := 1;
			Z := 0;
		end if;
		Pointp   := ETateWei!f1(ETate![X,Y,Z]);
		psi      := Isomorphism(ETateWei,Ep);
		Point    := psi(Pointp);

		Append(~Cocycles,J_E);
		Append(~Points,Point);
		print "X-coordinate : ", algdep(Point[1],2*#RedF);
		print "Y-coordinate : ", algdep(Point[2],2*#RedF);
	end for;
	
	return Points, Cocycles;
	
end function;


// ============================================================
// ===              6) Slope 1 invariants                   ===
// ============================================================

function Slope1ValueInfinity(J)

	K  := Parent(J[3]);
	Kp := Parent(J[4]);
	sqrtDFp := J[4];
	p  := J[6];
	m  := J[7];
	Fs := J[8];
	Gs := J[9];
	
	L := [];
	
	for i := 3 to Floor((m-1)/2) by 2 do
		Phi := [* [* Parent(Fs[1])!1, Fs[i] *], [* Parent(Fs[1])!1, Fs[i-1] *], J[3], J[4], J[5], J[6], J[7] *] ;
// 		Sub := [SubstitutionProduct(Phi[i],[IMat,SMat])  : i in [1,2]];	
// 		CS  := [constants_product(Sub[i],K,Kp,sqrtDFp,m) : i in [1,2]];
		a := CuspidalValue(Phi: tests := false);
		Append(~L,Log(a)/p^i);
	end for;
	
	b := L[#L];
	for i := 1 to #L do
		print i, Valuation(L[i]-b);
	end for;
	
	return L;	

end function;



function Slope1Invariant(J,G)

	p  := J[6];
	m  := J[7];
	Fs := J[8];
	Gs := J[9];
	
	L := [];
	
	for i := 3 to Floor((m-1)/2) by 2 do
		Phi := [* [* Parent(Fs[1])!1, Fs[i] *], J[2], J[3], J[4], J[5], J[6], J[7] *] ;
		a := RMValue(Phi,G);
		Append(~L,Log(a)/p^i);
	end for;
	
	b := L[#L];
	for i := 1 to #L do
		print i, Valuation(L[i]-b);
	end for;
	
	return b;	

end function;


// ============================================================
// ===                      7) Tests                        ===
// ============================================================


function RunTests()

	t0 := Cputime();
	
	// Test for D=32, p=17 (split) 
	Phi := StandardTest(D32F2,17,10);
	print "======================================"; 
	print " Tests D = 32, p = 17 (split) passed. ";
	
	// Test for D=12, p=3 (ramified, norm)
	Phi := StandardTest(D12F2,3,60 : nrm := true);  // This one is curious, look at level 0 constant in (1+S)-test. 
	print "======================================"; 
	print " Tests D = 12, p = 3 (ramified, norm) passed. ";
	
	// Test for D=5, p=2 (inert) 
	Phi := StandardTest(D5F1,2,200);
	print "======================================"; 
	print " Tests D = 5,  p = 2 (inert) passed. ";
	print "";
	print " --- Values at D=77.";
	print "Should satisfy 141*x^4 + 391*x^3 + 536*x^2 + 391*x + 141: ";
	a := RMValue(Phi,D77F1 : alg := true);
	a := RMValue(Phi,D77F2 : alg := true);
	print " --- Values at D=93.";
	print "Should satisfy 268879*x^4 - 128401*x^3 - 151356*x^2 - 128401*x + 268879: ";
	a := RMValue(Phi,D93F1 : alg := true);
	a := RMValue(Phi,D93F2 : alg := true);
	print " --- Gross-Stark unit: ";
	algdep(CuspidalValue(Phi : tests := true),2);
	
	// Test for D=321, p=7 (inert) 
	Phi := StandardTest(D321F2,7,100);
	print "======================================"; 
	print " Tests D = 321, p = 7 (inert) passed. ";
	print "";
	print " --- Gross-Stark unit: ";
	algdep(CuspidalValue(Phi : tests := true),6);
	print "";	
	
	return "All tests passed! Total time: ", Cputime()-t0; 

end function;



